using System;
using System.Collections.ObjectModel;
using System.Runtime.InteropServices;
using System.Text;
using System.IO.SensorServices.Rfid.Client;
using System.Threading;
using Microsoft.SensorServices.Rfid;

namespace Microsoft.RfidComWrapper
{
    /// <summary>
    /// RFID operations wrapper.
    /// </summary>
    [ComVisible(true)]
    [Guid("2E00C9EB-9400-450e-B6B8-F3BB6E4B4816"), ClassInterface(ClassInterfaceType.None)]
    [ComSourceInterfaces(typeof(IRfidWrapperEvents))]
    public class RfidWrapper : IRfidWrapper, IDisposable
    {
        private string errorMessage;
        private int errorCode;
        private bool disposed;
        private bool eventsTracking;

        /// <summary>
        /// Initializes a new instance of the <see cref="RfidWrapper"/> class with default parameters.
        /// Starts the thread that fires events when a new tag arrives.
        /// </summary>
        public RfidWrapper()
        {
            ThreadPool.QueueUserWorkItem(new WaitCallback(ReadRfidEvents));
        }

        /// <summary>
        /// Reads RFID tags from COM+ and fires events.
        /// </summary>
        private void ReadRfidEvents(object state)
        {
            while (!disposed)
            {
                if (eventsTracking)
                {
                    try
                    {
                        RfidTagWrapper tag;
                        using (RfidServicedWrapper srv = new RfidServicedWrapper())
                        {
                            tag = srv.ReadTag();
                        }

                        if (tag != null)
                        {
                            NewTagEventHandler handler = NewTag;
                            if (handler != null)
                            {
                                handler(tag);
                            }
                        }
                    }
                    catch (Exception ex)
                    {
                        HandleError(ex);
                    }
                }
                
                Thread.Sleep(500);
            }
        }

        #region Events
        /// <summary>
        /// A handler for the new tag event.
        /// </summary>
        [ComVisible(false)]
        public delegate void NewTagEventHandler(RfidTagWrapper tag);
        public event NewTagEventHandler NewTag;
        #endregion

        /// <summary>
        /// Writes a tag to a device.
        /// </summary>
        /// <param name="deviceName">The RFID Device name to connect to.</param>
        /// <param name="tagId">The tag ID to write.</param>
        /// <param name="tagData">The tag data to write.</param>
        public void WriteTag(string deviceName, string tagId, string tagData)
        {
            ResetState();
            
            ASCIIEncoding encode = new ASCIIEncoding();

            try
            {
                using (DeviceConnection Connection = new DeviceConnection(deviceName))
                {
                    Connection.Open();
                    Connection.PrintTag(deviceName, new PrintLabel(encode.GetBytes(tagId), encode.GetBytes(tagData), null, null, null, LockTargets.None), 1, null, null, null);
                    Connection.Close();
                }
            }
            catch (Exception ex)
            {
                HandleError(ex);
            }
        }

        /// <summary>
        /// Reads the tags count.
        /// </summary>
        /// <param name="deviceName">The name of the RFID device to read data from.</param>
        /// <returns>The tags count</returns>
        public int TagsCount(string deviceName)
        {
            ResetState();

            try
            {
                using (DeviceConnection Connection = new DeviceConnection(deviceName))
                {
                    //gets Tags
                    Connection.Open();
                    ReadOnlyCollection<TagReadEvent> tagList = (ReadOnlyCollection<TagReadEvent>)Connection.GetTags(deviceName, TagDataSelector.All, null, null);
                    Connection.Close();
                    return tagList.Count;
                }
            }
            catch (Exception ex)
            {
                HandleError(ex);
            }

            return 0;
        }

        /// <summary>
        /// Reads the tag data.
        /// </summary>
        /// <param name="deviceName">The name of the RFID device to read data from.</param>
        /// <param name="index">The index of the tag to read.</param>
        /// <returns>The tag data or null if the tag cannot be read. Check <see cref="ErrorCode"/> and <see cref="ErrorMessage"/> properties for the error description if any.</returns>
        public RfidTagWrapper ReadTag(string deviceName, int index)
        {
            ResetState();

            try
            {
                using (DeviceConnection Connection = new DeviceConnection(deviceName))
                {
                    //gets Tags
                    Connection.Open();
                    ReadOnlyCollection<TagReadEvent> tagList = (ReadOnlyCollection<TagReadEvent>)Connection.GetTags(deviceName, TagDataSelector.All, null, null);
                    Connection.Close();

                    if (index < 0 || index >= tagList.Count)
                    {
                        errorCode = 1;
                        errorMessage = "Tag index is invalid.";
                        return null;
                    }
                    
                    return RfidTagWrapper.CreateFromReadTagEvent(tagList[index]);
                }
            }
            catch (Exception ex)
            {
                HandleError(ex);
            }

            return null;
        }

        /// <summary>
        /// Gets the last error code.
        /// </summary>
        /// <value>An integer that represents the error code. 0 if no error is detected.</value>
        public int ErrorCode
        {
            get { return errorCode; }
        }

        /// <summary>
        /// Gets the last error message.
        /// </summary>
        /// <value>An string that represents the error description.</value>
        public string ErrorMessage
        {
            get { return errorMessage; }
        }

        /// <summary>
        /// Starts tracking RFID events.
        /// </summary>
        /// <returns>True if the operation has succeeded. False if an error has occured. Check <see cref="ErrorMessage"/> property for details.</returns>
        public bool StartEventsTracking()
        {
            ResetState();
            
            try
            {
                using (RfidServicedWrapper srv = new RfidServicedWrapper())
                {
                    srv.StartTagTracking();
                }
                eventsTracking = true;
                return true;
            }
            catch (Exception ex)
            {
                HandleError(ex);
            }

            return false;
        }

        /// <summary>
        /// Stops tracking events.
        /// </summary>
        /// <returns>True if operation has succeeded. False if an error has occured. Check <see cref="ErrorMessage"/> property for details.</returns>
        public bool StopEventsTracking()
        {
            ResetState();

            try
            {
                using (RfidServicedWrapper srv = new RfidServicedWrapper())
                {
                    srv.StopTagTracking();
                }
                eventsTracking = false;
                return true;
            }
            catch (Exception ex)
            {
                HandleError(ex);
            }

            return false;
        }

        /// <summary>
        /// Handles an error by setting needed properties.
        /// </summary>
        private void HandleError(Exception error)
        {
            errorCode = 1;
            errorMessage = error.Message;
        }
        
        /// <summary>
        /// Resets error state.
        /// </summary>
        private void ResetState()
        {
            errorCode = 0;
            errorMessage = string.Empty;
        }

        #region IRfidWrapperEvents Members

        #endregion

        #region IDisposable implementation
        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting managed and native resources.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Destructor logic.
        /// </summary>
        ~RfidWrapper()
        {
            Dispose(false);
        }
                
        /// <summary>
        /// The bulk of the clean-up code.
        /// </summary>
        /// <param name="disposing">If true, cleans up native and manged resources. If false, cleans up native resources only.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (!disposed)
            {
                disposed = true;
            }
        }

        #endregion
    }
}
